/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.time.Duration;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienAnfrager;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienException;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienSpeicherClient;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlAktualisierenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlAnlegenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlLesenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlLoeschenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSchreibenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAntwort;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherFehlerAntwort;

/**
 * Zentrale für die Kommunikation mit dem Ganglinienspeicher.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public final class DavGanglinienSpeicher implements GanglinienSpeicherClient {

    private static class AnfrageSpeicher {

        private final GlSpeicherAnfrage anfrage;
        private final Consumer<GlSpeicherAntwort> antwort;
        private final Consumer<GlSpeicherFehlerAntwort> fehlerAntwort;

        AnfrageSpeicher(GlSpeicherAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
                Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
            this.anfrage = anfrage;
            this.antwort = antwort;
            this.fehlerAntwort = fehlerAntwort;
        }
    }

    private DavGanglinieAnfrageSender sender;
    private ScheduledExecutorService timeOutChecker;

    private Map<String, AnfrageSpeicher> anfragen = Collections.synchronizedMap(new LinkedHashMap<>());
	private final GanglinienAnfrager anfrager;
	private final AtomicLong anfrageId = new AtomicLong(1);

    public DavGanglinienSpeicher(DavObjektFactory objektFactory) {
        this(objektFactory, objektFactory.getGanglinienSpeicherObjekt(null));
    }

    public DavGanglinienSpeicher(DavObjektFactory objektFactory, SystemObject speicherObj) {
        this.sender = new DavGanglinieAnfrageSender(objektFactory.getDav(), speicherObj);
        DavGanglinieAntwortEmpfaenger empfaenger = new DavGanglinieAntwortEmpfaenger(objektFactory.getDav(),
                objektFactory, speicherObj);
        empfaenger.antwortEmpfangen().addHandler(a -> ganglinienAntwortEmpfangen(a));
        empfaenger.fehlerAntwortEmpfangen().addHandler(a -> ganglinienFehlerAntwortEmpfangen(a));
       
        ClientApplication applicationObject = objektFactory.getDav().getLocalApplicationObject();
        anfrager = GanglinienAnfrager.of(applicationObject.getNameOrPidOrId(), applicationObject.getId());
        timeOutChecker = Executors.newScheduledThreadPool(10);
    }

	@Override
	public GanglinienAnfrager getAnfrager() {
		return anfrager;
	}
    
	@Override
	public String getAnfrageId() {
		return anfrager.getName() + "_" + anfrageId.getAndIncrement();
	}
	
    private void ganglinienFehlerAntwortEmpfangen(GlSpeicherFehlerAntwort a) {
        AnfrageSpeicher anfrageSpeicher = anfragen.remove(a.getAbsenderZeichen());
        if (anfrageSpeicher != null) {
            anfrageSpeicher.fehlerAntwort.accept(a);
        }
    }

    private void ganglinienAntwortEmpfangen(GlSpeicherAntwort a) {
        AnfrageSpeicher anfrageSpeicher = anfragen.remove(a.getAbsenderZeichen());
        if (anfrageSpeicher != null) {
            anfrageSpeicher.antwort.accept(a);
        }
    }

    @Override
    public List<MqGanglinie> liesGanglinien(GlLesenAnfrage anfrage, Duration timeout)
            throws TimeoutException, GanglinienException {

        AtomicReference<GlSpeicherAntwort> antwort = new AtomicReference<>(null);
        AtomicReference<GlSpeicherFehlerAntwort> fehler = new AtomicReference<>(null);

        try {
            final Semaphore warteAufAnwort = new Semaphore(1);
            warteAufAnwort.acquire();
            liesGanglinien(anfrage, a -> {
                antwort.set(a);
                warteAufAnwort.release();
            }, f -> {
                fehler.set(f);
                warteAufAnwort.release();
            });

            if (!warteAufAnwort.tryAcquire(timeout.toMillis(), TimeUnit.MILLISECONDS)) {
                throw new TimeoutException("Ganglinienspeicher hat nicht in der erwarteten Zeit geantwortet");
            }

        } catch (@SuppressWarnings("unused") InterruptedException e) {
            // ignore
        }

        if (antwort.get() == null) {
            if (fehler.get() != null) {
                throw new GanglinienException(fehler.get().getFehlerMeldungen().toString());
            }
            throw new TimeoutException("Ereigniskalender hat nicht in der erwarteten Zeit geantwortet");
        }

        return antwort.get().getGanglinien();
    }

    @Override
    public void liesGanglinien(GlLesenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
        anfragen.put(anfrage.getAbsenderZeichen(), new AnfrageSpeicher(anfrage, antwort, fehlerAntwort));
        sender.liesGanglinien(anfrage);
    }

    @Override
    public void liesGanglinien(GlLesenAnfrage anfrage, Consumer<GlSpeicherAntwort> speicherAntwort,
            Consumer<GlSpeicherFehlerAntwort> speicherFehlerAntwort, Duration timeout) {
        liesGanglinien(anfrage, speicherAntwort, speicherFehlerAntwort);
        timeOutChecker.schedule(() -> {
            AnfrageSpeicher anfrageSpeicher = anfragen.remove(anfrage.getAbsenderZeichen());
            if (anfrageSpeicher != null) {
                anfrageSpeicher.fehlerAntwort.accept(createTimeoutFehler(anfrageSpeicher));
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    private GlSpeicherFehlerAntwort createTimeoutFehler(AnfrageSpeicher anfrageSpeicher) {
        GlSpeicherFehlerAntwort fehlerAnwort = new GlSpeicherFehlerAntwort(
                anfrageSpeicher.anfrage.getAbsenderZeichen());
        fehlerAnwort.addFehlerMeldung("Timeout erreicht");
        fehlerAnwort.setMessQuerschnitt(anfrageSpeicher.anfrage.getMessQuerschnitt());
        return fehlerAnwort;
    }

    @Override
    public void schreibeGanglinien(GlSchreibenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
        anfragen.put(anfrage.getAbsenderZeichen(), new AnfrageSpeicher(anfrage, antwort, fehlerAntwort));
        sender.schreibeGanglinien(anfrage);
    }

    @Override
    public void schreibeGanglinien(GlSchreibenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort, Duration timeout) {
        schreibeGanglinien(anfrage, antwort, fehlerAntwort);
        timeOutChecker.schedule(() -> {
            AnfrageSpeicher anfrageSpeicher = anfragen.remove(anfrage.getAbsenderZeichen());
            if (anfrageSpeicher != null) {
                anfrageSpeicher.fehlerAntwort.accept(createTimeoutFehler(anfrageSpeicher));
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    @Override
    public void loescheGanglinien(GlLoeschenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
        anfragen.put(anfrage.getAbsenderZeichen(), new AnfrageSpeicher(anfrage, antwort, fehlerAntwort));
        sender.loescheGanglinien(anfrage);

    }

    @Override
    public void loescheGanglinien(GlLoeschenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort, Duration timeout) {
        loescheGanglinien(anfrage, antwort, fehlerAntwort);
        timeOutChecker.schedule(() -> {
            AnfrageSpeicher anfrageSpeicher = anfragen.remove(anfrage.getAbsenderZeichen());
            if (anfrageSpeicher != null) {
                anfrageSpeicher.fehlerAntwort.accept(createTimeoutFehler(anfrageSpeicher));
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    @Override
    public List<MqGanglinie> erzeugeGanglinien(GlAnlegenAnfrage anfrage, Duration timeout)
            throws TimeoutException, GanglinienException {
        AtomicReference<GlSpeicherAntwort> antwort = new AtomicReference<>(null);
        AtomicReference<GlSpeicherFehlerAntwort> fehler = new AtomicReference<>(null);

        try {
            final Semaphore warteAufAnwort = new Semaphore(1);
            warteAufAnwort.acquire();
            erzeugeGanglinien(anfrage, a -> {
                antwort.set(a);
                warteAufAnwort.release();
            }, f -> {
                fehler.set(f);
                warteAufAnwort.release();
            });

            if (!warteAufAnwort.tryAcquire(timeout.toMillis(), TimeUnit.MILLISECONDS)) {
                throw new TimeoutException("Ganglinienspeicher hat nicht in der erwarteten Zeit geantwortet");
            }

        } catch (@SuppressWarnings("unused") InterruptedException e) {
            // ignore
        }

        if (antwort.get() == null) {
            if (fehler.get() != null) {
                throw new GanglinienException(fehler.get().getFehlerMeldungen().toString());
            }
            throw new TimeoutException("Ganglinienspeicher hat nicht in der erwarteten Zeit geantwortet");
        }

        return antwort.get().getGanglinien();

    }

    @Override
    public void erzeugeGanglinien(GlAnlegenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
        anfragen.put(anfrage.getAbsenderZeichen(), new AnfrageSpeicher(anfrage, antwort, fehlerAntwort));
        sender.erzeugeGanglinien(anfrage);
    }

    @Override
    public void erzeugeGanglinien(GlAnlegenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort, Duration timeout) {
        erzeugeGanglinien(anfrage, antwort, fehlerAntwort);
        timeOutChecker.schedule(() -> {
            AnfrageSpeicher anfrageSpeicher = anfragen.remove(anfrage.getAbsenderZeichen());
            if (anfrageSpeicher != null) {
                anfrageSpeicher.fehlerAntwort.accept(createTimeoutFehler(anfrageSpeicher));
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    @Override
    public List<MqGanglinie> aktualisiereGanglinien(GlAktualisierenAnfrage anfrage, Duration timeout)
            throws TimeoutException, GanglinienException {
        AtomicReference<GlSpeicherAntwort> antwort = new AtomicReference<>(null);
        AtomicReference<GlSpeicherFehlerAntwort> fehler = new AtomicReference<>(null);

        try {
            final Semaphore warteAufAnwort = new Semaphore(1);
            warteAufAnwort.acquire();
            aktualisiereGanglinien(anfrage, a -> {
                antwort.set(a);
                warteAufAnwort.release();
            }, f -> {
                fehler.set(f);
                warteAufAnwort.release();
            });

            if (!warteAufAnwort.tryAcquire(timeout.toMillis(), TimeUnit.MILLISECONDS)) {
                throw new TimeoutException("Ganglinienspeicher hat nicht in der erwarteten Zeit geantwortet");
            }

        } catch (@SuppressWarnings("unused") InterruptedException e) {
            // ignore
        }

        if (antwort.get() == null) {
            if (fehler.get() != null) {
                throw new GanglinienException(fehler.get().getFehlerMeldungen().toString());
            }
            throw new TimeoutException("Ganglinienspeicher hat nicht in der erwarteten Zeit geantwortet");
        }

        return antwort.get().getGanglinien();
    }

    @Override
    public void aktualisiereGanglinien(GlAktualisierenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort) {
        anfragen.put(anfrage.getAbsenderZeichen(), new AnfrageSpeicher(anfrage, antwort, fehlerAntwort));
        sender.aktualisiereGanglinien(anfrage);
    }

    @Override
    public void aktualisiereGanglinien(GlAktualisierenAnfrage anfrage, Consumer<GlSpeicherAntwort> antwort,
            Consumer<GlSpeicherFehlerAntwort> fehlerAntwort, Duration timeout) {
        aktualisiereGanglinien(anfrage, antwort, fehlerAntwort);
        timeOutChecker.schedule(() -> {
            AnfrageSpeicher anfrageSpeicher = anfragen.remove(anfrage.getAbsenderZeichen());
            if (anfrageSpeicher != null) {
                anfrageSpeicher.fehlerAntwort.accept(createTimeoutFehler(anfrageSpeicher));
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

}
